import pandas as pd


class PandasExtra:
    def dict_to_df(self, dict_arr):
        df = pd.DataFrame.from_dict(dict_arr)
        return df

    def list_to_df(self, list_arr):
        return pd.DataFrame.from_records([s for s in list_arr])

    def help(self):
        print(f"=== {type(self).__name__} === \n")
        print(".dict_to_df(dict_arr) => convert dict in df\n")
        print(".list_to_df(list_arr) => convert list in df\n")



from __future__ import print_function
import pickle
import os.path
import pandas as pd
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
import gspread
from oauth2client.service_account import ServiceAccountCredentials


class GoogleSpreadsheet:
    def __init__(self, spreadsheet_id=None, sheet_name=None, credentials_json=None):
        """Scopes are various types of access(read,write,readonly etc.) specified to the Google API
        Incase of scope modification, delete the file token.pickle
        """
        try:
            assert spreadsheet_id is not None, "Please provide a spreadsheet id"
            assert sheet_name is not None, "Please provide a sheet name "
            assert (
                credentials_json is not None
            ), "Please provide credentials JSON file path"
        except AssertionError:
            print("Mising one or more arguments.Please provide all arguments")
            return

        # SpreadsheetID
        self.SPREADSHEET_ID = spreadsheet_id
        # Spreadsheet Name
        self.SHEET_NAME = sheet_name
        # Goodle drive credentials JSON file
        self.credentials_json = credentials_json

        self.SCOPES = ["https://www.googleapis.com/auth/spreadsheets.readonly"]
        # Establish connection to google sheet's API
        connection = self.get_connection()
        # Check if connection is established
        if connection:
            """If connection is positive
            Get data from spreadsheet
            """
            if not self.get_data(connection):
                print("Failed to get data from sheet")
        else:
            print("Failed to connect to Google Spreadsheet")

    def get_connection(self):
        """This Method uses the credentials.json file to establish connection with
        google spreadsheet API.
        """
        try:
            creds = None
            # The file token.pickle stores the user's access and refresh tokens, and is
            # created automatically when the authorization flow completes for the first
            # time.
            if os.path.exists("token.pickle"):
                with open("token.pickle", "rb") as token:
                    creds = pickle.load(token)
            # If there are no (valid) credentials available, let the user log in.
            if not creds or not creds.valid:
                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_config(
                        self.credentials_json, self.SCOPES
                    )
                    creds = flow.run_local_server(port=0)
                # Save the credentials for the next run
                with open("token.pickle", "wb") as token:
                    pickle.dump(creds, token)

            service = build("sheets", "v4", credentials=creds)

            # Call the Sheets API
            sheet = service.spreadsheets()
            return sheet

        except Exception as e:
            print(e)
            return False

    def get_data(self, sheet):
        try:
            result = (
                sheet.values()
                .get(spreadsheetId=self.SPREADSHEET_ID, range=self.SHEET_NAME)
                .execute()
            )
            # values = result.get('values', [])

            # Assumes first line is header!
            header = result.get("values", [])[0]
            values = result.get("values", [])[1:]  # Everything else is data.
            if not values:
                print("No data found.")
            else:
                all_data = []
                for col_id, col_name in enumerate(header):
                    column_data = []
                    for row in values:
                        column_data.append(row[col_id])
                    ds = pd.Series(data=column_data, name=col_name)
                    all_data.append(ds)
                df = pd.concat(all_data, axis=1)
                print("Dataframe:\n", df)
                return True
        except Exception as e:
            print(e)
            return False

    def get_google_sheet(self, title, sheet_name):
        scope = [
            "https://spreadsheets.google.com/feeds",
            "https://www.googleapis.com/auth/drive",
        ]
        creds = ServiceAccountCredentials.from_json_keyfile_name(
            "client_secret.json", scope
        )
        client = gspread.authorize(creds)
        book = client.open(title)
        sheets = book.worksheets()

        for sheet in sheets:
            if sheet.title == sheet_name:
                return sheet.get_all_records()
        return False



import pandas as pd
from geopy.extra.rate_limiter import RateLimiter


class GeoLocator:
    """GeoLocator allows users to fetch latitude and longitude for
    a list of address in dataframe column
    Example usage:
        geo = GeoLocator(limit=50)
        geo.use_google(api_key="your_api_key") #to set the service to google maps
         OR
        geo.use_osm(user_agent="your_app_name") #to set the service to open street maps
        df = geo.geoloc(df, column="address")
    """

    def __init__(self, limit=50):
        self.client = None
        self.limit = limit

    def use_google(
        self,
        api_key: str,
        domain: str = "maps.googleapis.com",
        min_delay_seconds: int = 1,
        **kwargs,
    ):
        from geopy.geocoders import GoogleV3

        client = GoogleV3(api_key=api_key, domain=domain, **kwargs)
        self.client = RateLimiter(client.geocode, min_delay_seconds=min_delay_seconds)

    def use_osm(self, user_agent: str, min_delay_seconds: int = 1, **kwargs):
        if min_delay_seconds < 1:
            raise ValueError("Minimum 1 second delay is required for osm")
        from geopy.geocoders import Nominatim

        client = Nominatim(user_agent=user_agent, **kwargs)
        self.client = RateLimiter(client.geocode, min_delay_seconds=min_delay_seconds)

    def geoloc(self, df: pd.DataFrame, column: str) -> pd.DataFrame:
        if not self.client:
            raise ValueError(
                "Please set the map service using either `use_google` or `use_osm` method"
            )
        if df.shape[0] > self.limit:
            raise ValueError(
                f"Dataset number of rows is more than the set limit of {self.limit}"
            )
        df["_location"] = df[column].apply(self.client)
        df["LATITUDE"] = df["_location"].apply(
            lambda loc: loc.latitude if loc else None
        )
        df["LONGITUDE"] = df["_location"].apply(
            lambda loc: loc.longitude if loc else None
        )
        df = df.drop(columns=["_location"])
        return df

    def help(self):
        print(f"=== {type(self).__name__} === \n")
        print(
            '.use_google(api_key, domain: str = "maps.googleapis.com", min_delay_seconds: int = 1, **kwargs)\
                => get file from ftp path\n'
        )
        print(".geoloc(df, column) => get file from ftp path\n")
        print(
            ".use_osm(user_agent, min_delay_seconds, kwargs) => do ls in ftp in path\n"
        )



from requests.auth import HTTPBasicAuth
from escapism import escape
import requests
import string
import base64
import json
import os

_docker_safe_chars = set(string.ascii_letters + string.digits)
_docker_escape_char = "-"


def _escape(s):
    """Escape a string to docker-safe characters"""
    return escape(
        s,
        safe=_docker_safe_chars,
        escape_char=_docker_escape_char,
    )


class Me:
    base_public_url = None
    endpoint = "me"
    auth = None
    req_headers = {"Accept": "application/json", "Content-Type": "application/json"}

    def __init__(self, base_url, auth, endpoint="me"):
        self.base_public_url = base_url
        self.auth = auth
        self.endpoint = endpoint

    def get(self):
        req = requests.get(
            url=f"{self.base_public_url}/{self.endpoint}",
            headers=self.req_headers,
            auth=self.auth,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def update(self, data):
        req = requests.put(
            url=f"{self.base_public_url}/{self.endpoint}",
            headers=self.req_headers,
            auth=self.auth,
            json=data,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def help(self):
        print("=== Me === \n")
        print(".get() => get my user data\n")
        print(
            ".update(data) => update my user data (only few fields are allow if your are not admin)\n"
        )
        print(".delete(data) => delete one data by id (should be in the doc)\n")


class CRUD:
    base_public_url = None
    endpoint = None
    auth = None
    req_headers = {"Accept": "application/json", "Content-Type": "application/json"}

    def __init__(self, base_url, endpoint, auth):
        self.base_public_url = base_url
        self.endpoint = endpoint
        self.auth = auth

    def get_all(self, search=None, sort=None, limit=20, skip=0):
        params = {
            "limit": limit,
            "skip": skip,
        }
        if search:
            params["search"] = json.dumps(search)
        if sort:
            params["sort"] = json.dumps(sort)
        req = requests.get(
            url=f"{self.base_public_url}/{self.endpoint}",
            params=params,
            headers=self.req_headers,
            auth=self.auth,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def get(self, id):
        req = requests.get(
            url=f"{self.base_public_url}/{self.endpoint}/{id}",
            headers=self.req_headers,
            auth=self.auth,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def insert(self, data):
        req = requests.post(
            url=f"{self.base_public_url}/{self.endpoint}",
            auth=self.auth,
            headers=self.req_headers,
            json=data,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def update(self, data):
        id = data.get("_id")
        req = requests.put(
            url=f"{self.base_public_url}/{self.endpoint}/{id}",
            auth=self.auth,
            headers=self.req_headers,
            json=data,
            allow_redirects=False,
        )
        req.raise_for_status()

    def delete(self, data):
        id = data.get("_id")
        req = requests.delete(
            url=f"{self.base_public_url}/{self.endpoint}/{id}",
            auth=self.auth,
            headers=self.req_headers,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def help(self):
        print("=== CRUD === \n")
        print(
            ".get_all(search=None, sort=None, limit=20, skip=0) => get all data in collection with optional filters\n"
        )
        print(".get(id) => get one data by id\n")
        print(".insert(data) => insert one data\n")
        print(".update(data) => update one data by id (should be in the doc)\n")
        print(".delete(data) => delete one data by id (should be in the doc)\n")


class SmartTable(CRUD):
    database = None
    collection = None

    def __init__(self, base_url, database, collection, auth):
        self.database = database
        self.collection = collection
        endpoint = f"smarttables/{database}/{collection}"
        CRUD.__init__(self, base_url, endpoint, auth)

    def allowed(self, database, collection):
        req = requests.post(
            url=f"{self.base_public_url}/{self.endpoint}/allowed",
            auth=self.auth,
            headers=self.req_headers,
            allow_redirects=False,
        )
        req.raise_for_status()
        req_json = req.json()
        return True if req_json.allowed else False

    def delete_all(self):
        req = requests.delete(
            url=f"{self.base_public_url}/{self.endpoint}/all",
            auth=self.auth,
            headers=self.req_headers,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def help(self):
        print("=== Internals === \n")
        print(
            ".allowed(database, collection) => check if you are allowed to get data from this database and collection\n"
        )
        print(".delete_all() => delete all data in a collection\n")
        super().help()


class DarkKnight:
    """BOB lib"""

    base_public_url = None
    smart_tables = []
    __auth = None
    req_headers = {"Accept": "application/json", "Content-Type": "application/json"}

    def __init__(self, api_key=None, user=None, PUBLIC_DK_API=None):
        """
        Description: This class connect you to a darknight instance
        """
        self.user = user if user else os.environ.get("JUPYTERHUB_USER", user)
        self.base_public_url = (
            PUBLIC_DK_API
            if PUBLIC_DK_API
            else os.environ.get("PUBLIC_DK_API", PUBLIC_DK_API)
        )
        self.__auth = HTTPBasicAuth(self.user, api_key)
        self.users = CRUD(self.base_public_url, "users", self.__auth)
        self.workspaces = CRUD(self.base_public_url, "workspaces", self.__auth)
        self.me = Me(self.base_public_url, self.__auth)

    def get_notebook_public_url(self, token=""):

        client = self.user
        clientEncoded = _escape(client)
        message_bytes = clientEncoded.encode("ascii")
        base64_bytes = base64.b64encode(message_bytes)
        username_base64 = base64_bytes.decode("ascii")
        return f"{self.base_public_url}/notebook/{username_base64}/{token}"

    def init_smarttable(self, database, collection):
        return SmartTable(self.base_public_url, database, collection, self.__auth)

    def notification(self, email, subject, content, image_url, link_url):
        data = {
            "email": email,
            "object": subject,
            "content": content,
            "image_url": image_url,
            "link_url": link_url,
        }
        req = requests.post(
            url=f"{self.base_public_url}/notifications/send",
            auth=self.__auth,
            headers=self.req_headers,
            json=data,
            allow_redirects=False,
        )
        req.raise_for_status()
        return req

    def help(self):
        print("=== DarkKnight === \n")
        print(
            ".init_smarttable(database, collection) => initialise smart table with database and collection\n"
        )
        print(
            ".init_smarttable(None, None).help() => show the helper of init_smarttable class\n"
        )
        print(".users.help() => show the helper of user class\n")
        print(".workspaces.help() => show the helper of workspaces class\n")
        print(".me.help() => show the helper of me class\n")



import pandas as pd
from pandas.tseries.offsets import Day


class Prediction:
    def __init__(self):

        self.param_model_map = {"arima": "ARIMA", "linear": "LINEAR", "svr": "SVR"}

        # either all to predict using all the models else one of arima, svr or linear
        self.prediction_type = None

        # input data
        self.input_df = None

        # the column name to be used as the date column
        self.date_column = None

        # The value to be predicted
        self.label = None

        # the number of data points in the future to be predicted, max 1000
        self.data_points = 20

        # True if the final output needs to be stored as graph, False otherwise
        self.plot = True

        self.model_params = dict(
            arima=dict(
                start_p=1,
                start_q=1,
                test="adf",  # use adftest to find optimal 'd'
                max_p=3,
                max_q=3,  # maximum p and q
                m=1,  # frequency of series
                d=None,  # let model determine 'd'
                seasonal=False,  # No Seasonality / True for seasoanal i.e. SARIMA
                start_P=0,
                D=0,
                trace=True,
                error_action="ignore",
                suppress_warnings=True,
                stepwise=True,
            ),
            linear=dict(fit_intercept=True, normalize=False, n_jobs=None),
            svr=dict(kernel="rbf", C=1e3, gamma=0.1),
        )
        self.final_predicted_df = None
        self.plot_df = None

    def plot_output(
        self, data_points: int = 0, plot_width: int = 15, plot_height: int = 6
    ):
        return self.__plot_output(data_points, plot_width, plot_height)

    def init_class_vars(
        self,
        prediction_type: str,
        dataset: pd.DataFrame,
        label: str,
        date_column: str,
        data_points: int,
        plot: bool,
    ):

        # either all to predict using all the models else one of arima, svr or linear
        self.prediction_type = prediction_type

        # input data
        if isinstance(dataset, pd.Series):
            self.input_df = dataset.to_frame()
        elif isinstance(dataset, pd.DataFrame):
            self.input_df = dataset
        else:
            raise ValueError("The dataset should either be a Dataframe or Series.")

        # the column name to be used as the date column
        self.date_column = date_column

        # The value to be predicted
        if label in self.input_df.columns:
            self.label = label
        else:
            raise ValueError("The label should a column name of the dataset.")

        # the number of data points in the future to be predicted, max 1000
        self.data_points = data_points

        # True if the final output needs to be stored as graph, False otherwise
        self.plot = plot

    def __createmodel(self, model_type):

        if model_type == "ARIMA":
            import pmdarima as pm

            model = pm.auto_arima(
                self.input_df[self.label], **self.model_params["arima"]
            )
            predicted_values = model.predict(n_periods=self.data_points)
        elif model_type in ("LINEAR", "SVR"):
            prediction = (
                self.input_df[self.label]
                .shift(-self.data_points)
                .to_numpy()
                .reshape(-1, 1)
            )
            X = self.input_df[self.label].to_numpy().reshape(-1, 1)[: -self.data_points]
            y = prediction[: -self.data_points].ravel()
            if model_type == "SVR":
                from sklearn.svm import SVR

                model = SVR(**self.model_params["svr"])
            else:
                from sklearn.linear_model import LinearRegression

                model = LinearRegression(**self.model_params["linear"])
            model.fit(X, y)
            X_predict = (
                self.input_df[self.label]
                .to_numpy()
                .reshape(-1, 1)[-self.data_points :]  # noqa: E203
            )
            predicted_values = model.predict(X_predict)
        else:
            raise ValueError(
                "Please specify an prediction_type as arima OR linear OR svr or all"
            )
        return model, predicted_values

    def __transform_output(self, data_df, predicted_values, predicted_col):
        predicted_index = pd.date_range(
            start=data_df.index.max() + Day(1), periods=self.data_points
        )
        predict_df = pd.DataFrame(
            data=predicted_values, index=predicted_index, columns=[predicted_col]
        )
        predict_df.index.name = self.date_column
        return predict_df

    def __modelling_prediction(self):
        if self.prediction_type == "all":
            output_dfs = [
                self.input_df,
            ]
            models_dict = {}
            predicted_cols = []
            for param in self.model_params:
                model_type = self.param_model_map[param]

                predicted_col = model_type
                predicted_cols.append(predicted_col)
                model, predicted_values = self.__createmodel(model_type)
                models_dict[model_type] = model
                predicted_df = self.__transform_output(
                    self.input_df.copy(), predicted_values, predicted_col
                )
                output_dfs.append(predicted_df)
        else:
            output_dfs = [
                self.input_df,
            ]
            models_dict = {}
            predicted_cols = []
            model_type = self.param_model_map[self.prediction_type]
            predicted_col = model_type
            predicted_cols.append(predicted_col)
            model, predicted_values = self.__createmodel(model_type)
            models_dict[model_type] = model
            predicted_df = self.__transform_output(
                self.input_df.copy(), predicted_values, predicted_col
            )
            output_dfs.append(predicted_df)
        return predicted_cols, output_dfs

    def __plot_output(
        self, data_points: int, plot_width: int = 15, plot_height: int = 6
    ):
        if data_points == 0:
            data_points = self.data_points + 7
        if self.plot:
            if self.prediction_type == "all":
                # lines no from the df = no of prediction_types * the datapoints required
                line_nos = data_points * 5  # lines from the end to plot
            else:
                line_nos = data_points * 3  # lines from the end to plot

            return self.plot_df.tail(line_nos).plot.line(
                figsize=(plot_width, plot_height)
            )
        else:
            raise ValueError("Rerun the prediction with plot set to True")

    def __melt_output(self, predicted_cols, output_dfs):
        final_df = pd.concat(
            output_dfs,
            ignore_index=False,
            axis=1,
        )
        for col in predicted_cols:
            final_df[col] = final_df[col].fillna(value=self.input_df[self.label])
        final_df["COMPOUND"] = final_df[predicted_cols].mean(axis=1)

        final_df.index.name = self.date_column
        final_df = final_df.rename(columns={self.label: "ACTUAL"})
        plot_df = final_df.copy()
        final_df = final_df.reset_index().melt(
            id_vars=[self.date_column], var_name="LABEL", value_name="VALUE"
        )
        final_df = final_df.sort_values(by=[self.date_column, "LABEL"])

        return final_df, plot_df

    def predict(
        self,
        prediction_type: str,
        dataset: pd.DataFrame,
        label: str,
        date_column: str,
        data_points: int = 20,
        plot: bool = True,
    ):
        # initializes the class variables
        self.init_class_vars(
            prediction_type=prediction_type,
            dataset=dataset,
            label=label,
            date_column=date_column,
            data_points=data_points,
            plot=plot,
        )

        # modelling and making the predictions
        predicted_cols, output_dfs = self.__modelling_prediction()

        # making the output more user friendly and also plotting the graph
        self.final_predicted_df, self.plot_df = self.__melt_output(
            predicted_cols, output_dfs
        )
        return self.final_predicted_df
